
#include "lowe.h"

QedQcd::QedQcd()
  : a(2), mf(9), mtPole(PMTOP) { 
  setPars(11);
  // Default object: 1998 PDB defined in 'def.h'
  mf(1) = MUP; mf(2) = MCHARM; 
  mf(4) = MDOWN; mf(5) = MSTRANGE; mf(6) = MBOTTOM;
  mf(7) = MELECTRON; mf(8) = MMUON; mf(9) = MTAU;
  a(1) = 7.8180608e-3;  a(2) = ALPHASMZ;
  mf(3) = getRunMtFromMz(PMTOP, ALPHASMZ);
  setMu(MZ);
  setLoops(3);
  setThresholds(1);
}

const QedQcd & QedQcd::operator=(const QedQcd & m) {
  if (this == &m) return *this;
  mtPole = m.mtPole;
  a = m.a;
  mf = m.mf;
  setLoops(m.displayLoops());
  setThresholds(m.displayThresholds());
  return *this;
}

//For communication with outside routines: sets all data by one vector y=1..11.
void QedQcd::set(const DoubleVector & y) {
  a(ALPHA) = y.display(1);
  a(ALPHAS) = y.display(2);
  int i; for (i=3; i<=11; i++)
    mf(i-2) = y.display(i);
}

DoubleVector QedQcd::display() const {
  DoubleVector y(11);
  y(1) = a.display(ALPHA);
  y(2) = a.display(ALPHAS);
  int i; for (i=3; i<=11; i++)
    y(i) = mf.display(i-2);
  return y;
}

//  Active flavours at energy mu
int QedQcd::flavours(double mu) const {
  int k = 0;
  if (mu > mf.display(mTop)) k++;
  if (mu > mf.display(mCharm)) k++;
  if (mu > mf.display(mUp)) k++;
  if (mu > mf.display(mDown)) k++;
  if (mu > mf.display(mBottom)) k++;
  if (mu > mf.display(mStrange)) k++;
  return k;
}

ostream & operator <<(ostream &left, const QedQcd &m) {
  left << "mU: " << m.displayMass(mUp) 
       << "  mC: " << m.displayMass(mCharm) 
       << "  mt: " << m.displayMass(mTop) 
       << "  mt^pole: " << m.displayPoleMt()
       << endl;
  left << "mD: " << m.displayMass(mDown) 
       << "  mS: " << m.displayMass(mStrange)
       << "  mB: " << m.displayMass(mBottom) << endl;
  left << "mE: " << m.displayMass(mElectron) 
       << "  mM: " << m.displayMass(mMuon) 
       <<  "  mT: " << m.displayMass(mTau) << endl;
  left << "aE: " << m.displayAlpha(ALPHA) << 
    "  aS: " << m.displayAlpha(ALPHAS)  
       << "  scale: " << m.displayMu() << endl;
  left << "loops: " << m.displayLoops() 
       << "          thresh: " << m.displayThresholds() << endl;
  return left;
}

istream & operator >>(istream &left, QedQcd &m) {
  char c[12];
  double mu, mc, mtpole, md, ms, mb, me, mmu, mtau, alph, alphas, scale;
  int t, l;
  left >> c >> mu >> c >> mc >> c >> c >> c >>
    mtpole;
  left >> c >> md >> c >> ms >> c >> mb;
  left >> c >> me >> c >> mmu >> c >> mtau;
  left >> c >> alph >> c >> alphas >> c >> scale;
  left  >> c >> l >> c >> t;
  m.setMass(mUp, mu);
  m.setMass(mCharm, mc);
  m.setMass(mDown, md);
  m.setMass(mStrange, ms);
  m.setMass(mBottom, mb);
  m.setMass(mElectron, me);
  m.setMass(mMuon, mmu);
  m.setMass(mTau, mtau);
  m.setAlpha(ALPHA, alph);
  m.setAlpha(ALPHAS, alphas);
  m.setMu(scale);
  // y[3] is pole mass
  m.setPoleMt(mtpole);
  m.setMass(mTop, getRunMtFromMz(mtpole, alphas));
  // default 3-loop qcd calculation
  m.setLoops(l);
  m.setThresholds(t);

  return left;
}

//  returns qed beta function at energy mu < mtop
double QedQcd::qedBeta() const {
  double x;
  x = 24.0 / 9.0;
  if (displayMu() > mf.display(mCharm)) x += 8.0 / 9.0;
  if (displayMu() > mf.display(mTop)) x += 8.0 / 9.0;
  if (displayMu() > mf.display(mBottom)) x += 2.0 / 9.0;
  if (displayMu() > mf.display(mTau)) x += 2.0 / 3.0;
  if (displayMu() > MZ) x += 1.0 / 6.0;
  if (displayMu() > (mtPole + EPS))  {
      cerr << "qed beta function called at " << displayMu() << 
	"above mt=" << displayPoleMt() << 
	", outside range of validity";
      cerr << " in QedQcd::qedbeta\n";
      exit(1);
    }
  
  return (x * sqr(a.display(ALPHA)) / PI);
}

//  next routine calculates beta function to 3 loops in qcd for The Standard
//  Model. Note that if quark masses are running, the number of active quarks
//  will take this into account. Returns beta
double QedQcd::qcdBeta() const {
  int quarkFlavours = flavours(this->displayMu());
  static const double INVPI = 1.0 / PI;
  double qb0, qb1, qb2;
  qb0 = (11.0e0 - (2.0e0 / 3.0e0 * quarkFlavours)) / 4;
  qb1 = (102.0e0 - (38.0e0 * quarkFlavours) / 3.0e0) / 16;
  qb2 = (2.857e3 / 2 - (5.033e3 * quarkFlavours) / 1.8e1  + 
 	 (3.25e2 * sqr(quarkFlavours) ) / 5.4e1) / 64; 
  double qa0, qa1, qa2;
  
  if (displayLoops() < 0 || displayLoops() > 3) {
      cout << "Wrong loops parameter :" << displayLoops() << " in " << *this;
      exit(1);
    }
  if (displayLoops() > 0) qa0 = qb0 * INVPI; else qa0 = 0.0;
  if (displayLoops() > 1) qa1 = qb1 * sqr(INVPI); else qa1 = 0.0;
  if (displayLoops() > 2) qa2 = qb2 * sqr(INVPI) * INVPI; else qa2 = 0.0;
  
  // add contributions of the one, two and three loop constributions resp.
  double beta;
  beta = -2 * sqr(displayAlpha(ALPHAS)) * 
    (qa0 + qa1 * displayAlpha(ALPHAS) + qa2 *
     sqr(displayAlpha(ALPHAS)));
  return beta;
}

//(See comments for above function). returns a vector x(1..9) of fermion mass
//beta functions -- been checked!
void QedQcd::massBeta(DoubleVector & x) const {
  static const double INVPI = 1.0 / PI, ZETA3 = 1.202056903159594;
  
  // qcd bits: 1,2,3 loop resp.
  double qg1, qg2, qg3;
  int quarkFlavours = flavours(displayMu());
  qg1 = INVPI;
  qg2 = (202.0 / 3.0 - (20.0e0 * quarkFlavours) / 9.0) * sqr(INVPI) / 16;
  qg3 = (1.249e3 - ((2.216e3 * quarkFlavours) / 27.0e0 - 
		    (1.6e2 * ZETA3 * quarkFlavours) / 3.0e0) -
	 140.0e0 * quarkFlavours * quarkFlavours / 81.0e0) * sqr(INVPI) *
    INVPI / 64;
  
  double qcd = -2.0 * a.display(ALPHAS) * (qg1  + qg2 * a.display(ALPHAS) +
					   qg3 * sqr(a.display(ALPHAS)));
  double qed = -a.display(ALPHA) * INVPI / 2;
  
  int i;
  for (i=1;i<=3;i++)   // up quarks
    x(i) = (qcd + 4.0 * qed / 3.0) * mf.display(i);
  for (i=4;i<=6;i++)   // down quarks
    x(i) = (qcd + qed / 3.0) * mf.display(i);
  for (i=7;i<=9;i++)   // leptons
    x(i) = 3.0 * qed * mf.display(i);
  // switch off relevant beta functions
  if (displayThresholds() > 0)
    for(i=1;i<=9;i++) if (displayMu() < displayMass()(i)) x(i) = 0.0;
}

DoubleVector QedQcd::beta() const {
  DoubleVector dydx(11);
  dydx(1) = qedBeta();
  dydx(2) = qcdBeta();
  DoubleVector y(9);
  massBeta(y);
  int i; for (i=3; i<=11; i++)
    dydx(i) = y(i-2);
  return dydx;
}

void QedQcd::runGauge(double x1, double x2) {
  DoubleVector y(2);
  tempLe = this;
  y(1) = tempLe->displayAlpha(ALPHA);
  y(2) = tempLe->displayAlpha(ALPHAS);
  
  callRK(x1, x2, y, gaugeDerivs);
  
  setAlpha(ALPHA, y(1));
  setAlpha(ALPHAS, y(2));
}

// Takes QedQcd object created at MZ and spits it out at mt
void QedQcd::toMt() {
  double alphasMZ = displayAlpha(ALPHAS);
  double alphaMZ = displayAlpha(ALPHA);
  runGauge(MZ, 1.0);
  //Run whole lot up to pole top mass
  double mt = this->displayPoleMt();
  run(1.0, MZ);
  // Reset alphas to erase numerical integration errors.
  setAlpha(ALPHAS, alphasMZ);
  setAlpha(ALPHA, alphaMZ);
  run(MZ, mt);
}

// Takes QedQcd object created at MZ and spits it out at MZ
void QedQcd::toMz() {
  double alphasMZ = displayAlpha(ALPHAS);
  double alphaMZ = displayAlpha(ALPHA);
  runGauge(MZ, 1.0);
  run(1.0, MZ);
  // Reset alphas to erase numerical integration errors.
  setAlpha(ALPHAS, alphasMZ);
  setAlpha(ALPHA, alphaMZ);
}


// This will calculate the three gauge couplings of the Standard Model at the
// scale m2.
// It's a simple one-loop calculation only and no
// thresholds are assumed. Range of validity is electroweak to top scale.
// alpha1 is in the GUT normalisation. sinth = sin^2 thetaW(Q) in MSbar
// scheme
DoubleVector QedQcd::getGaugeMu(const double m2, const double sinth) const {

  if (displayMu() != MZ) {
    cout << "Error in lowe.cpp:QedQcd::getGaugeMu called with mu=" <<
      displayMu() << " not MZ=" << MZ << endl;
    exit(1);
    }

  DoubleVector temp(1, 3);

  double a1, a2, aem = displayAlpha(ALPHA), m1 = displayMu();
  // Set alpha1,2 at scale m1 from data:
  a1 = 5.0 * aem / (3.0 * (1.0 - sinth));
  a2 = aem / sinth;
  static const double INVPI = 1.0 / PI;

  // Renormalise a1,a2 to scale m2 assuming topless SM with one light Higgs
  // doublet
  a1 = 1.0 / ( 1.0 / a1 + 4.0 * INVPI * 1.07e2 * log(m1 / m2) / 2.4e2 );
  a2 = 1.0 / ( 1.0 / a2 - 4.0 * INVPI * 2.50e1 * log(m1 / m2) / 4.8e1 );

  temp.set(1, a1);
  temp.set(2, a2);
  // calculate alphas(m2)
  QedQcd oneset(*this); oneset.runto(oneset.displayPoleMt());
  // Set alphas(m) to be what's already calculated.
  temp.set(3, oneset.displayAlpha(ALPHAS));

  return temp;
}

int accessedReadIn; // Should be initialised to zero at start of prog
/*
--------------- read in a qcd-type object ------------------
Call with fname "" if you want it to come from standard input

data initialisation file: for lowERge
mu,c,t(pole)
md,s,b
me,mu,tau -- all in MSbar running masses unless stated
alpha(MSbar), alphas(MSbar)
mu
*/
void readIn(QedQcd &mset, char fname[80]) {
  static QedQcd prevReadIn; // Data will be stored in here for rest of the
				// run

  // Read in data if it's not been set
  if (accessedReadIn == 0) {
    if (!strcmp(fname,"")) cin >> prevReadIn; // from standard input 
    else {   
      // read from filename fname
	  fstream fin(fname, ios::in); 
	  if(!fin) {
	    cerr << "Can't find input file " << fname << endl;
	    exit(1);
	  }
	  fin >> prevReadIn;
	  fin.close();
    }
    
    cout << prevReadIn;
    accessedReadIn = 1; // Flag the fact we've read in the data once
  }

  mset = prevReadIn;
}

DoubleVector gaugeDerivs(double x, const DoubleVector & y) {
  tempLe->setMu(exp(x));
  tempLe->setAlpha(ALPHA, y.display(1));
  tempLe->setAlpha(ALPHAS, y.display(2));
  DoubleVector dydx(2);
  dydx(1) = tempLe->qedBeta();
  dydx(2) = tempLe->qcdBeta();    
  return dydx;
}

// Given pole mass and alphaS(MZ), returns running top mass -- one loop qcd
double getRunMtFromMz(double poleMt, double asMZ) {
  return getRunMt(poleMt, getAsmt(poleMt, asMZ));
}

// Input pole mass of top and alphaS(mt), outputs running mass mt(mt)
// including one-loop standard model correction only
double getRunMt(double poleMt, double asmt) {
  return poleMt / (1.0 + (4.0 / (3.0 * PI)) * asmt);
}

// Given a value of mt, and alphas(MZ), find alphas(mt) to 1 loops in qcd:
// it's a very good approximation at these scales, better than 10^-3 accuracy
double getAsmt(double mtop, double alphasMz) {
  return alphasMz / (1.0 - 23.0 * alphasMz / (6.0 * PI) * log(MZ / mtop));
}

